from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Optional, Union

from .attr import AttrId

if TYPE_CHECKING:
    from .drepr import DRepr


@dataclass
class AlignedStep:
    source_idx: int
    target_idx: int


@dataclass
class RangeAlignment:
    source: AttrId
    target: AttrId
    aligned_steps: list[AlignedStep]

    def swap(self) -> RangeAlignment:
        return RangeAlignment(
            self.target,
            self.source,
            [AlignedStep(s.target_idx, s.source_idx) for s in self.aligned_steps],
        )

    def compute_cardinality(self, desc: DRepr) -> Cardinality:
        """
        Compute the cardinality of an alignment

        The cardinality between attribute `x` and attribute `y` are defined as follows:

        1. one-to-one: one item of `x` can only link to one item of `y` and vice versa.
        2. one-to-many: one item of `x` can link to multiple items of `y`, but one item of `y` can only
           link to one item of `x`.
        3. many-to-one: the reversed case of one-to-many
        4. many-to-many: multiple items of `x` can link to multiple items of `y` and vice versa.

        The cardinality depends on the number of unfixed dimensions of each attribute, if an attribute
        has no unfixed steps, it will be one-to-*, otherwise many-to-*
        """
        source = desc.get_attr_by_id(self.source)
        target = desc.get_attr_by_id(self.target)

        source_nary_steps = set(source.path.get_nary_steps())
        target_nary_steps = set(target.path.get_nary_steps())
        for step in self.aligned_steps:
            source_nary_steps.discard(step.source_idx)
            target_nary_steps.discard(step.target_idx)

        if len(source_nary_steps) == 0:
            if len(target_nary_steps) == 0:
                return Cardinality.OneToOne
            else:
                return Cardinality.OneToMany
        else:
            if len(target_nary_steps) == 0:
                return Cardinality.ManyToOne
            else:
                return Cardinality.ManyToMany


@dataclass
class ValueAlignment:
    source: AttrId
    target: AttrId

    def swap(self) -> ValueAlignment:
        return ValueAlignment(self.target, self.source)

    def compute_cardinality(self, desc: DRepr) -> Cardinality:
        """
        Compute the cardinality of an alignment

        The cardinality between attribute `x` and attribute `y` are defined as follows:

        1. one-to-one: one item of `x` can only link to one item of `y` and vice versa.
        2. one-to-many: one item of `x` can link to multiple items of `y`, but one item of `y` can only
           link to one item of `x`.
        3. many-to-one: the reversed case of one-to-many
        4. many-to-many: multiple items of `x` can link to multiple items of `y` and vice versa.

        The cardinality of the join will be one-to-* or *-to-one if values of source & target are unique.
        """
        source = desc.get_attr_by_id(self.source)
        target = desc.get_attr_by_id(self.target)

        if source.unique:
            if target.unique:
                return Cardinality.OneToOne
            else:
                return Cardinality.OneToMany
        else:
            if target.unique:
                return Cardinality.ManyToOne
            else:
                return Cardinality.ManyToMany


@dataclass
class IdenticalAlign:
    source: AttrId
    target: AttrId

    def compute_cardinality(self, desc: DRepr) -> Cardinality:
        return Cardinality.OneToOne

    def swap(self) -> IdenticalAlign:
        return IdenticalAlign(self.target, self.source)


@dataclass
class AutoAlignment:
    attrs: Optional[list[str]] = None

    def compute_cardinality(self, desc: DRepr) -> Cardinality:
        raise NotImplementedError(
            """Auto alignment will be preprocessed to turn into Range/Value alignment, so it should not be used to compute cardinality"""
        )


Alignment = Union[RangeAlignment, ValueAlignment, IdenticalAlign, AutoAlignment]


class AlignmentType(Enum):
    Range = "range"
    Value = "value"
    Ident = "identical"
    Auto = "auto"


class Cardinality(Enum):
    OneToOne = "one-to-one"
    OneToMany = "one-to-many"
    ManyToOne = "many-to-one"
    ManyToMany = "many-to-many"

    def is_star_to_many(self) -> bool:
        return self == Cardinality.OneToMany or self == Cardinality.ManyToMany

    def is_one_to_star(self) -> bool:
        return self == Cardinality.OneToOne or self == Cardinality.OneToMany

    def is_many_to_star(self) -> bool:
        return self == Cardinality.ManyToOne or self == Cardinality.ManyToMany



from __future__ import annotations

from pathlib import Path

import openpyxl


class RASpreadsheet:

    def __init__(self, workbook: openpyxl.Workbook):
        self.workbook = workbook

    def __getitem__(self, name: str):
        return self.workbook[name]


def read_source_spreadsheet(infile: Path | str):
    wb = openpyxl.load_workbook(infile)
    return RASpreadsheet(wb)



import re
from abc import ABC, abstractmethod
from copy import copy
from typing import List, Optional, Union

from drepr.models.parse_v1.path_parser import PathParser
from drepr.utils.validator import InputError

from ..path import Expr, IndexExpr, Path, RangeExpr, WildcardExpr
from ..resource import Resource, ResourceType


class PathParserV2(PathParser):
    """
    Similar to path parser v1. However, we now allow special using indexing name per resource type

    For spreadsheet, we allow column to be letter
    """

    REG_SRANGE = re.compile(r"^(\d+)?\.\.(-?\d+)?(?::(\d+))?$")
    REG_SINDEX = re.compile(r"^(?:\$\{([^}]+)})|(\d+)|(.*)$")
    REG_SRANGE_EXPR = re.compile(
        r"^(?:(\d+)|(?:\$\{([^}]+)}))?\.\.(?:(-\d+)|(?:\$\{([^}]+)}))?(?::(\d+)|(?:\$\{([^}]+)}))?$"
    )

    REG_JPATH_BRACKET = re.compile(
        r"(?:\[(-?\d+)?\:(-?\d+)?(?:\:(-?\d+))?\])|(?:\[(-?\d+)\])|(?:\['([^']+)'\])"
    )
    REG_JPATH_DOT = re.compile(r"\.((?:(?!\.|\[).)+)")

    def parse(
        self, resource: Optional[Resource], path: Union[str, list], parse_trace: str
    ) -> Path:
        if isinstance(path, str):
            return self.parse_jsonpath(resource, path, parse_trace)

        if isinstance(path, list):
            return self.parse_custom_path(resource, path, parse_trace)

        raise InputError(
            f"{parse_trace}\nERROR: the path must either be a "
            f"string (JSONPath) or a list of steps. Get {type(path)} instead"
        )

    # noinspection PyMethodMayBeStatic
    def letter2index(self, letter: str) -> int:
        letter = list(letter.lower())
        n_chars = ord("z") - ord("a") + 1
        index = 0
        for i, c in enumerate(reversed(letter)):
            assert (
                ord("a") <= ord(c) <= ord("z")
            ), f"{c} is not a valid column in spreadsheet"
            index += (ord(c) - ord("a") + 1) * (n_chars**i)
        return index - 1

    def isdigit(self, s: str) -> bool:
        if s.startswith("-"):
            return s[1:].isdigit()
        return s.isdigit()

    def parse_jsonpath(
        self, resource: Optional[Resource], jpath: str, parse_trace: str
    ) -> Path:
        if not jpath.startswith("$"):
            raise InputError(
                f"{parse_trace}\nERROR: invalid json path. The path must start with `$`. "
                f"Get: {jpath}"
            )

        jpath = jpath[1:]
        steps = []
        parsing_pos = 1

        # pre-processing the spreadsheet resource to allow letter column
        if resource is not None:
            if resource.type == ResourceType.Spreadsheet:
                last_step_index = max(jpath.rfind("["), jpath.rfind("."))
                if jpath[last_step_index] == "[":
                    last_step = jpath[last_step_index + 1 : -1]
                    result = last_step.split(":")
                    if len(result) == 1:
                        index = result[0]
                        if not self.isdigit(index):
                            new_last_step = self.letter2index(index)
                        else:
                            new_last_step = index
                    else:
                        if len(result) == 3:
                            start, end, step = result
                        elif len(result) == 2:
                            start, end = result
                            step = 1
                        else:
                            raise InputError(f"{parse_trace}\nERROR: invalid path")

                        if (len(start) > 0 and not self.isdigit(start)) or (
                            len(end) > 0 and not self.isdigit(end)
                        ):
                            # they use letter system, otherwise, do nothing
                            if (len(start) > 0 and self.isdigit(start)) or (
                                len(end) > 0 and self.isdigit(end)
                            ):
                                raise InputError(
                                    f"{parse_trace}\nERROR: Cannot mixed between number and letter index"
                                )
                            start = self.letter2index(start)
                            if len(end) > 0:
                                end = self.letter2index(end)
                            new_last_step = f"{start}:{end}:{step}"
                        else:
                            new_last_step = f"{start}:{end}:{step}"

                    jpath = jpath[:last_step_index] + f"[{new_last_step}]"
                elif jpath[last_step_index] == ".":
                    last_step = jpath[last_step_index + 1 :]
                    if not self.isdigit(last_step):
                        new_last_step = self.letter2index(last_step)
                    else:
                        new_last_step = last_step
                    jpath = jpath[:last_step_index] + f".{new_last_step}"

        while len(jpath) > 0:
            if jpath.startswith("["):
                m = self.REG_JPATH_BRACKET.match(jpath)
                if m is None:
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid json path, error while parsing bracket at position {parsing_pos}"
                    )

                jpath = jpath[m.span()[-1] :]
                parsing_pos += m.span()[-1]  # m.span()[0] is always 0

                if m.group(5) is not None:
                    # match with string
                    steps.append(IndexExpr(m.group(5)))
                elif m.group(4) is not None:
                    # match with a single number
                    steps.append(IndexExpr(int(m.group(4))))
                else:
                    steps.append(
                        RangeExpr(
                            int(m.group(1) or "0"),
                            int(m.group(2)) if m.group(2) is not None else None,
                            int(m.group(3) or "1"),
                        )
                    )
            elif jpath.startswith(".*~"):
                # *~ select property names
                steps.append(WildcardExpr.Names)
                jpath = jpath[3:]
                parsing_pos += 3
            elif jpath.startswith(".*"):
                steps.append(WildcardExpr.Values)
                jpath = jpath[2:]
                parsing_pos += 2
            else:
                m = self.REG_JPATH_DOT.match(jpath)
                if m is None:
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid json path, error while parsing step at position {parsing_pos}"
                    )

                jpath = jpath[m.span()[-1] :]
                parsing_pos += m.span()[-1]  # m.span()[0] is always 0

                # after a dot, it can either be a number or a string
                if m.group(1).isdigit():
                    steps.append(IndexExpr(int(m.group(1))))
                else:
                    steps.append(IndexExpr(m.group(1)))

        return Path(steps)

    def parse_custom_path(
        self, resource: Optional[Resource], path: List[str], parse_trace: str
    ) -> Path:
        if resource is not None:
            if resource.type == ResourceType.Spreadsheet:
                path = copy(path)
                last_step = path[-1]
                if isinstance(last_step, str):
                    if last_step.find("..") != -1:
                        tmp = last_step.split(":")
                        start, end = tmp[0].split("..")
                        if len(tmp) == 2:
                            step = f":{tmp[1]}"
                        else:
                            step = ""

                        if (len(start) > 0 and not self.isdigit(start)) or (
                            len(end) > 0 and not self.isdigit(end)
                        ):
                            # they use letter system, otherwise, do nothing
                            if (len(start) > 0 and self.isdigit(start)) or (
                                len(end) > 0 and self.isdigit(end)
                            ):
                                raise InputError(
                                    f"{parse_trace}\nERROR: Cannot mixed between number and letter index"
                                )
                            start = self.letter2index(start)
                            if len(end) > 0:
                                end = self.letter2index(end)
                            new_last_step = f"{start}..{end}{step}"
                        else:
                            new_last_step = f"{start}..{end}{step}"
                        path[-1] = new_last_step
                    elif not self.isdigit(last_step):
                        path[-1] = self.letter2index(last_step)

        steps = []
        for i, step in enumerate(path):
            trace = f"Parsing step {i} ({step})"
            if isinstance(step, str):
                m = self.REG_SRANGE.match(step)
                if m is not None:
                    steps.append(
                        RangeExpr(
                            int(m.group(1) or "0"),
                            int(m.group(2)) if m.group(2) is not None else None,
                            int(m.group(3) or "1"),
                        )
                    )
                    continue

                m = self.REG_SRANGE_EXPR.match(step)
                if m is not None:
                    steps.append(
                        RangeExpr(
                            (
                                (
                                    Expr(m.group(1)[2:-1])
                                    if m.group(1).startswith("${")
                                    else int(m.group(1))
                                )
                                if m.group(1) is not None
                                else 0
                            ),
                            (
                                (
                                    Expr(m.group(2)[2:-1])
                                    if m.group(2).startswith("${")
                                    else int(m.group(2))
                                )
                                if m.group(2) is not None
                                else None
                            ),
                            (
                                (
                                    Expr(m.group(2)[2:-1])
                                    if m.group(2).startswith("${")
                                    else int(m.group(2))
                                )
                                if m.group(2) is not None
                                else 1
                            ),
                        )
                    )
                    continue

                if step.startswith("${"):
                    steps.append(IndexExpr(Expr(step[2:-1])))
                else:
                    steps.append(IndexExpr(step))
            elif isinstance(step, int):
                steps.append(IndexExpr(step))
            elif isinstance(step, list):
                assert (
                    len(step) == 1
                ), "The list notation [...] is used to annotate optional index, so the list must have only one element"
                steps.append(IndexExpr(step[0], is_optional=True))
            else:
                raise InputError(
                    f"{parse_trace}\n{trace}\nERROR: step must either be string or number. Get {type(step)} instead"
                )

        return Path(steps)



from __future__ import annotations

from dataclasses import dataclass

from codegen.models import PredefinedFn, Program, Var, expr

from drepr.models.resource import ResourceType


class DReprPredefinedFn(PredefinedFn):

    @dataclass
    class _safe_item_getter(expr.Expr):
        collection: expr.Expr
        item: expr.Expr
        msg: expr.Expr

        def to_python(self):
            return f"safe_item_getter({self.collection.to_python()}, {self.item.to_python()}, {self.msg.to_python()})"

    @dataclass
    class _safe_len(expr.Expr):
        collection: expr.Expr
        msg: expr.Expr

        def to_python(self):
            return f"safe_len({self.collection.to_python()}, {self.msg.to_python()})"

    @staticmethod
    def safe_item_getter(
        program: Program,
        collection: expr.Expr,
        item: expr.Expr,
        msg: expr.Expr,
    ):
        program.import_("drepr.utils.safe.safe_item_getter", True)
        return DReprPredefinedFn._safe_item_getter(collection, item, msg)

    @staticmethod
    def safe_len(program: Program, collection: expr.Expr, msg: expr.Expr):
        program.import_("drepr.utils.safe.safe_len", True)
        return DReprPredefinedFn._safe_len(collection, msg)

    @staticmethod
    def read_source(program: Program, source_type: ResourceType, input_file: Var):
        program.import_(f"drepr.readers.prelude.read_source_{source_type.value}", True)
        return expr.ExprFuncCall(
            expr.ExprIdent(f"read_source_{source_type.value}"),
            [expr.ExprVar(input_file)],
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

import drepr.models.path as dpath
from drepr.models.parse_v2.path_parser import PathParserV2


class Container:

    def __init__(self, attrs: Optional[dict[str, AttributeData]] = None):
        self.attr = attrs or {}

    def __setitem__(self, id: str, attr: AttributeData):
        assert id not in self.attr
        self.attr[id] = attr

    def __getitem__(self, id: str) -> AttributeData:
        return self.attr[id]


@dataclass
class AttributeData:
    """Storing data of an attribute in the same location as in its origin resource.

    Since we cannot pre-populate the size of the attribute data, it has to built dynamically, so we need to store
    the data metadata along for rebuilding
    """

    steps: list[dpath.StepExpr]
    current_step_index: int
    value_step_index: int
    # can be an attribute data or real data (anything but it)
    value: dict[int | str, AttributeData | Any]

    @staticmethod
    def from_raw_path(raw_path: list):
        path = PathParserV2().parse(None, raw_path, "")
        attrdata = AttributeData.from_path(path.steps)
        assert attrdata is not None
        return attrdata

    @staticmethod
    def from_path(steps: list[dpath.StepExpr], idx: int = 0):
        if idx == len(steps):
            return None

        next_idx = idx
        while next_idx < len(steps) - 1:
            step = steps[next_idx]
            if not isinstance(step, dpath.IndexExpr):
                break

            if isinstance(step.val, dpath.Expr):
                # I don't know about recursive path expression yet -- what usecase and how to use them -- so I can't implement.
                raise Exception(
                    f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.val}"
                )

            next_idx += 1

        step = steps[next_idx]
        if isinstance(step, dpath.IndexExpr):
            if isinstance(step.val, dpath.Expr):
                # I don't know about recursive path expression yet -- what usecase and how to use them -- so I can't implement.
                raise Exception(
                    f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.val}"
                )

            # this is the end of the step, we just store it in a dictionary for simplicity (one-item dict)
            # this simply the implementation although we can do better
            return AttributeData(steps, idx, next_idx, {step.val: None})

        if isinstance(step, dpath.SetIndexExpr):
            # we know the keys, so we can pre-populate the data
            value = {}
            for val in step.vals:
                if isinstance(val, dpath.Expr):
                    # I don't know about recursive path expression yet -- what usecase and how to use them -- so I can't implement.
                    raise Exception(
                        f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.vals}"
                    )
                value[val] = AttributeData.from_path(steps, next_idx + 1)
            return AttributeData(steps, idx, next_idx, value)

        if isinstance(step, dpath.RangeExpr):
            if isinstance(step.start, dpath.Expr):
                # I don't know about recursive path expression yet -- what usecase and how to use them -- so I can't implement.
                raise Exception(
                    f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.start}"
                )
            if step.start > 0:
                return AttributeData(
                    steps, idx, next_idx, {i: None for i in range(step.start)}
                )

        return AttributeData(steps, idx, next_idx, {})

    def __len__(self):
        if self.current_step_index == self.value_step_index:
            return len(self.value)
        assert self.current_step_index < self.value_step_index
        return 1

    def __getitem__(self, key: int | str) -> AttributeData | Any:
        if self.current_step_index == self.value_step_index:
            if key not in self.value:
                if self.current_step_index == len(self.steps) - 1:
                    # this is the last one
                    self.value[key] = None
                else:
                    next_value_step_index = self._next_non_index_step_idx(
                        self.steps, self.value_step_index + 1
                    )
                    if next_value_step_index == len(self.steps):
                        # the rest of them are all index steps
                        self.value[key] = AttributeData(
                            self.steps,
                            self.current_step_index + 1,
                            len(self.steps) - 1,
                            {},
                        )
                    else:
                        self.value[key] = AttributeData(
                            self.steps,
                            self.current_step_index + 1,
                            next_value_step_index,
                            {},
                        )
            return self.value[key]
        assert self.current_step_index < self.value_step_index
        current_step = self.steps[self.current_step_index]
        assert isinstance(current_step, dpath.IndexExpr) and current_step.val == key
        return AttributeData(
            self.steps,
            self.current_step_index + 1,
            self.value_step_index,
            self.value,
        )

    def __setitem__(self, key: int | str, val: Any):
        if self.current_step_index == self.value_step_index:
            self.value[key] = val
            return
        assert self.current_step_index < self.value_step_index
        current_step = self.steps[self.current_step_index]
        assert isinstance(current_step, dpath.IndexExpr) and current_step.val == key

    @staticmethod
    def _next_non_index_step_idx(steps: list[dpath.StepExpr], start: int = 0):
        for i in range(start, len(steps)):
            if not isinstance(steps[i], dpath.IndexExpr):
                return i
        return len(steps)



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, TypeAlias, Union

from .path import Path
from .resource import ResourceId


class Sorted(Enum):
    Null = "none"
    Ascending = "ascending"
    Descending = "descending"


class ValueType(Enum):
    UnspecifiedSingle = "unspecified-single"
    Int = "int"
    Float = "float"
    String = "str"
    List_Int = "list[int]"
    List_Float = "list[float]"
    List_String = "list[str]"

    def is_list(self) -> bool:
        return self in (
            ValueType.List_Int,
            ValueType.List_Float,
            ValueType.List_String,
        )


MISSING_VALUE_TYPE = Optional[Union[str, int, float]]
AttrId: TypeAlias = str


@dataclass
class Attr:
    id: AttrId
    resource_id: ResourceId
    path: Path
    missing_values: list[MISSING_VALUE_TYPE]
    unique: bool = False
    sorted: Sorted = Sorted.Null
    value_type: ValueType = ValueType.UnspecifiedSingle

    @staticmethod
    def deserialize(raw: dict) -> "Attr":
        attr = Attr(**raw)
        attr.path = Path.deserialize(raw["path"])
        attr.sorted = Sorted(attr.sorted)
        attr.value_type = ValueType(attr.value_type)

        return attr



from __future__ import annotations

from collections import defaultdict
from dataclasses import asdict
from typing import TYPE_CHECKING, List

from drepr.models.parse_v2.path_parser import PathParserV2
from drepr.utils.validator import *

from ..align import AlignmentType, RangeAlignment
from ..parse_v1.align_parser import AlignParser
from ..parse_v1.attr_parser import AttrParser
from ..parse_v1.preprocessing_parser import PreprocessingParser
from ..parse_v1.resource_parser import ResourceParser
from ..sm import ClassNode, DataNode, LiteralNode, SemanticModel
from .sm_parser import SMParser

if TYPE_CHECKING:
    from ..drepr import DRepr


class ReprV2Parser:
    """
    The D-REPR language version 2 has similar to the schema of the first version.

    Difference with previous features:
    1. For spreadsheet columns, they can the letter instead of number
    2. Semantic model configuration is changed to focus on classes
    """

    TOP_KEYWORDS = {
        "version",
        "resources",
        "preprocessing",
        "attributes",
        "alignments",
        "semantic_model",
    }
    DEFAULT_RESOURCE_ID = "default"

    @classmethod
    def parse(cls, raw: dict):
        from ..drepr import DRepr

        Validator.must_be_subset(
            cls.TOP_KEYWORDS,
            raw.keys(),
            setname="Keys of D-REPR configuration",
            error_msg="Parsing D-REPR configuration",
        )

        for prop in ["version", "resources", "attributes"]:
            Validator.must_have(raw, prop, error_msg="Parsing D-REPR configuration")

        Validator.must_equal(
            raw["version"], "2", "Parsing D-REPR configuration version"
        )
        resources = ResourceParser.parse(raw["resources"])

        if len(resources) == 1:
            default_resource_id = resources[0].id
        else:
            default_resource_id = ResourceParser.DEFAULT_RESOURCE_ID

        path_parser = PathParserV2()
        preprocessing = PreprocessingParser(path_parser).parse(
            default_resource_id, resources, raw.get("preprocessing", [])
        )
        attrs = AttrParser(path_parser).parse(
            default_resource_id, resources, raw["attributes"]
        )
        aligns = AlignParser.parse(raw.get("alignments", []))

        if "semantic_model" in raw:
            sm = SMParser.parse(raw["semantic_model"])
            sm.prefixes.update(SemanticModel.get_default_prefixes())
        else:
            sm = None

        return DRepr(resources, preprocessing, attrs, aligns, sm)

    @classmethod
    def dump(cls, drepr: "DRepr", simplify: bool = True, use_json_path: bool = False):
        version = "2"
        sm = OrderedDict()

        class_counter = defaultdict(int)
        class_ids: Dict[str, str] = {}
        for node in drepr.sm.nodes.values():
            if isinstance(node, ClassNode):
                class_counter[node.label] += 1
                class_ids[node.node_id] = f"{node.label}:{class_counter[node.label]}"
                sm[class_ids[node.node_id]] = OrderedDict(
                    [
                        ("properties", []),
                        ("static_properties", []),
                        ("inverse_static_properties", []),
                        ("links", []),
                    ]
                )

        for node in drepr.sm.nodes.values():
            if isinstance(node, DataNode):
                for edge in [
                    e for e in drepr.sm.edges.values() if e.target_id == node.node_id
                ]:
                    if node.data_type is not None:
                        prop = (edge.label, node.attr_id, node.data_type.get_rel_uri())
                    else:
                        prop = (edge.label, node.attr_id)
                    sm[class_ids[edge.source_id]]["properties"].append(prop)

            if isinstance(node, LiteralNode):
                for edge in [
                    e for e in drepr.sm.edges.values() if e.target_id == node.node_id
                ]:
                    if node.data_type is not None:
                        prop = (edge.label, node.value, node.data_type.get_rel_uri())
                    else:
                        prop = (edge.label, node.value)
                    sm[class_ids[edge.source_id]]["static_properties"].append(prop)
                for edge in [
                    e for e in drepr.sm.edges.values() if e.source_id == node.node_id
                ]:
                    if edge.target_id not in class_ids:
                        raise Exception(
                            "D-Repr YAML version 2 does not support link from literal node to non-class nodes"
                        )
                    sm[class_ids[edge.target_id]]["inverse_static_properties"].append(
                        (edge.label, node.value)
                    )

        for edge in drepr.sm.edges.values():
            if isinstance(drepr.sm.nodes[edge.source_id], ClassNode) and isinstance(
                drepr.sm.nodes[edge.target_id], ClassNode
            ):
                sm[class_ids[edge.source_id]]["links"].append(
                    (edge.label, class_ids[edge.target_id])
                )
            if edge.is_subject:
                v = drepr.sm.nodes[edge.target_id]
                assert isinstance(v, DataNode)
                sm[class_ids[edge.source_id]]["subject"] = v.attr_id

        sm["prefixes"] = drepr.sm.prefixes

        preprocessing: List[dict] = []
        for prepro in drepr.preprocessing:
            preprocessing.append(OrderedDict([("type", prepro.type.value)]))
            for k, v in asdict(prepro.value).items():
                preprocessing[-1][k] = v
            preprocessing[-1]["path"] = prepro.value.path.to_lang_format(use_json_path)

        return OrderedDict(
            [
                ("version", version),
                (
                    "resources",
                    OrderedDict(
                        [
                            (
                                res.id,
                                OrderedDict(
                                    [("type", res.type.value)]
                                    + (
                                        [(k, v) for k, v in asdict(res.prop).items()]
                                        if res.prop is not None
                                        else []
                                    )
                                ),
                            )
                            for res in drepr.resources
                        ]
                    ),
                ),
                ("preprocessing", preprocessing),
                (
                    "attributes",
                    OrderedDict(
                        [
                            (
                                attr.id,
                                OrderedDict(
                                    [
                                        ("resource_id", attr.resource_id),
                                        (
                                            "path",
                                            attr.path.to_lang_format(use_json_path),
                                        ),
                                        ("unique", attr.unique),
                                        ("sorted", attr.sorted.value),
                                        ("value_type", attr.value_type.value),
                                        ("missing_values", attr.missing_values),
                                    ]
                                ),
                            )
                            for attr in drepr.attrs
                        ]
                    ),
                ),
                (
                    "alignments",
                    [
                        (
                            OrderedDict(
                                [
                                    ("type", AlignmentType.Range.value),
                                    ("source", align.source),
                                    ("target", align.target),
                                    (
                                        "aligned_dims",
                                        [
                                            OrderedDict(
                                                [
                                                    ("source", step.source_idx),
                                                    ("target", step.target_idx),
                                                ]
                                            )
                                            for step in align.aligned_steps
                                        ],
                                    ),
                                ]
                            )
                            if isinstance(align, RangeAlignment)
                            else OrderedDict(
                                [
                                    ("type", AlignmentType.Value.value),
                                    ("source", align.source),
                                    ("target", align.target),
                                ]
                            )
                        )
                        for align in drepr.aligns
                    ],
                ),
                ("semantic_model", sm),
            ]
        )



from __future__ import annotations

from dataclasses import dataclass

from drepr.models.sm import ClassNode, SemanticModel


@dataclass
class ReversedTopoSortResult:
    # list of class node ids
    topo_order: list[str]
    # list of edge ids that are removed
    removed_outgoing_edges: dict[int, bool]


def topological_sorting(sm: SemanticModel) -> ReversedTopoSortResult:
    """
    suppose we have a semantic model, that may contains circle, our task is to
    generate a topological order from the directed cyclic graph, and a set of edges
    involved in the cycles that are removed in order to generate the topo-order
    """
    visited_nodes = {uid: False for uid in sm.nodes}
    tmp_visited_nodes = {uid: False for uid in sm.nodes}
    removed_outgoing_edges = {eid: False for eid in sm.edges}

    while True:
        # pick a node from the remaining nodes and do reverse dfs. if we found a cycle,
        # then we break the cycle and repeat the process until we no cycle left
        random_start_node = next(iter(sm.nodes))
        has_unvisited_node = False
        for uid, u in sm.nodes.items():
            if not visited_nodes[uid] and isinstance(u, ClassNode):
                random_start_node = uid
                has_unvisited_node = True
                break

        if not has_unvisited_node:
            # we don't have any unvisited node left, so we finish!
            break

        # loop until it breaks all cycles
        while dfs_breaking_cycle(sm, random_start_node, [], removed_outgoing_edges):
            pass

        # mark visited nodes so that we can just skip them
        reverse_dfs(sm, random_start_node, visited_nodes, removed_outgoing_edges)

    # now we get acyclic graph, determine the topo-order
    reversed_topo_order = []
    for uid in sm.nodes:
        visited_nodes[uid] = False
        tmp_visited_nodes[uid] = False

    for uid, u in sm.nodes.items():
        if not visited_nodes[uid] and isinstance(u, ClassNode):
            dfs_reverse_topo_sort(
                sm,
                reversed_topo_order,
                uid,
                visited_nodes,
                tmp_visited_nodes,
                removed_outgoing_edges,
            )

    return ReversedTopoSortResult(
        topo_order=reversed_topo_order, removed_outgoing_edges=removed_outgoing_edges
    )


def dfs_reverse_topo_sort(
    sm: SemanticModel,
    topo_order: list[str],
    node: str,
    visited_nodes: dict[str, bool],
    tmp_visited_nodes: dict[str, bool],
    removed_outgoing_edges: dict[int, bool],
):
    """
    Generate a topological order of class nodes in the semantic model. The graph must be acyclic
    before using this function

    Based on DFS algorithm in here: https://en.wikipedia.org/wiki/Topological_sorting
    """
    if visited_nodes[node]:
        return

    if tmp_visited_nodes[node]:
        raise Exception("The graph has cycle!")

    tmp_visited_nodes[node] = True

    for e in sm.iter_outgoing_edges(node):
        if not removed_outgoing_edges[e.edge_id] and isinstance(
            sm.nodes[e.target_id], ClassNode
        ):
            dfs_reverse_topo_sort(
                sm,
                topo_order,
                e.target_id,
                visited_nodes,
                tmp_visited_nodes,
                removed_outgoing_edges,
            )

    tmp_visited_nodes[node] = False
    visited_nodes[node] = True
    topo_order.append(node)


def dfs_breaking_cycle(
    sm: SemanticModel,
    node: str,
    visited_path: list[str],
    removed_outgoing_edges: dict[int, bool],
) -> bool:
    """
    Try to break cycles using invert DFS. It returns true when break one cycle, and it terminates
    immediately. Thus, requires you to run this function many times until it return false
    """
    visited_path.append(node)

    for e in sm.iter_incoming_edges(node):
        if not removed_outgoing_edges[e.edge_id]:
            if e.source_id in visited_path:
                # this node is visited before in the path, and it is visited by traveling through `e`, we can drop `e` and move on
                removed_outgoing_edges[e.edge_id] = True
                return True

            if dfs_breaking_cycle(
                sm, e.source_id, visited_path, removed_outgoing_edges
            ):
                return True

    visited_path.pop()
    return False


def reverse_dfs(
    sm: SemanticModel,
    node: str,
    visited_nodes: dict[str, bool],
    removed_outgoing_edges: dict[int, bool],
) -> bool:
    """Reverse DFS to visit all ancestors of a node"""
    visited_nodes[node] = True
    for e in sm.iter_incoming_edges(node):
        if not removed_outgoing_edges[e.edge_id]:
            reverse_dfs(sm, e.source_id, visited_nodes, removed_outgoing_edges)



from __future__ import annotations

from typing import Any


class StreamClassWriter:

    def has_written_record(self, class_id: str, subject: str) -> bool:
        """Test if a record of a given class `class_id` has been written"""
        ...

    def begin_record(self, subject: str, is_blank: bool) -> bool:
        """
        Tell the writer that we are going to write a record, and we already have all information of the
        record to write.
        The function return `true` if this is a new record, otherwise `false`.

        # Arguments

        * `subject` - real id of the record, which is determined by the value of `drepr:uri` property
                                of the record or generated randomly (if it's a blank node)
        * `is_blank` - whether this subject is a blank node or not
        """
        ...

    def end_record(self):
        """
        Tell the writer that we finish writing all information of the record. This method cannot be
        called before `begin_record` method.

        Note: this method should not be called if the function `begin_record` return `false`
        """
        ...

    def begin_partial_buffering_record(self, subject: str, is_blank: bool) -> bool:
        """
        Tell the writer that we are going to write a record, and we don't have some information about
        the links between the records and other records as the other records haven't been generated
        yet.
        The function return `true` if this is a new record, otherwise `false`.

        # Arguments

        * `subject` - real id of the record, which is determined by the value of `drepr:uri` property
                                of the record
        * `is_blank` - whether this subject is a blank node or not
        """
        ...

    def write_data_property(self, subject: str, predicate_id: str, value: Any):
        """
        Write value of a data property of the current record.

        # Arguments

        * `subject` - real id of the current record
        * `predicate_id` - id of the data property that we are writing to (`edge_id` of the edge in the
                        semantic model)
        * `value` - value of the property
        """

    def write_object_property(
        self,
        predicate_id: str,
        object: str,
        is_subject_blank: bool,
        is_object_blank: bool,
        is_new_subj: bool,
    ):
        """
        Write value of a object property of the current record.

        # Arguments

        * `predicate_id` - id of the object property that we are writing to (`edge_id` of the edge in the
                        semantic model)
        * `object` - id of the target record
        * `is_subject_blank` - whether the subject id is blank
        * `is_object_blank` - whether the object id is blank
        * `is_new_subj` - whether the current record is a new record or an existing record (obtained
                          from the `begin_record` or `begin_partial_buffering_record` function
        """
        ...

    def buffer_object_property(
        self,
        target_cls: str,
        predicate_id: str,
        object: str,
        is_object_blank: bool,
    ):
        """
        Write value of a object property of the current record into buffer because we haven't generated
        the target object, it may be missed later.

        # Arguments

        * `target_cls`: id of the class of the target record that the current record is linked to,
                        which is the `node_id` of the class node in the semantic model
        * `predicate_id` - id of the object property that we are writing to (`edge_id` of the edge in the
                        semantic model)
        * `object` - id of the target record
        * `is_object_blank` - whether the object is blank node or not
        """



from __future__ import annotations

from typing import Any, Callable, Literal, Optional, Protocol

import drepr.models.path as path
from codegen.models import AST, DeferredVar, PredefinedFn, Program, Var, expr, stmt
from drepr.models.align import IdenticalAlign
from drepr.models.prelude import Alignment, Attr, DRepr, RangeAlignment
from drepr.program_generation.predefined_fn import DReprPredefinedFn
from drepr.program_generation.program_space import VarSpace


class AlignmentFn:

    def __init__(self, desc: DRepr, program: Program):
        self.desc = desc
        self.program = program

    def align(
        self,
        ast: AST,
        aligns: list[Alignment],
        validate_path: bool,
        on_missing_key: Optional[Callable[[AST], Any]],
        iter_final_list: bool,
    ):
        for align in aligns:
            if isinstance(align, RangeAlignment):
                ast = self.align_by_range(
                    ast, align, validate_path, on_missing_key, iter_final_list
                )
            elif isinstance(align, IdenticalAlign):
                # if this is the alignment between the same attribute -- do nothing
                if align.source == align.target:
                    continue

                # the two attributes are different, but they have identical align, it means that they are single-value attributes
                # we just need to access the value of the target attribute
                target = self.desc.get_attr_by_id(align.target)
                assert all(
                    isinstance(step, path.IndexExpr) for step in target.path.steps
                )
                ast = PathAccessor(self.program).iterate_elements(
                    ast,
                    target,
                    aligned_attr=None,
                    to_aligned_dim=None,
                    validate_path=validate_path,
                    on_missing_key=on_missing_key,
                    iter_final_list=iter_final_list,
                )
            else:
                raise NotImplementedError(type(align))
        return ast

    def align_by_range(
        self,
        ast: AST,
        align: RangeAlignment,
        validate_path: bool,
        on_missing_key: Optional[Callable[[AST], Any]],
        iter_final_list: bool,
    ) -> AST:
        """Generate a piece of code that will generate variables (of target attr) to
        complete this alignment, if the alignment is one/many to one, then return ast is the same
        as we do not introduce nested statements. If the alignment is one/many to many, then
        we will need to have a for loop, hence, we have nested statements -> nested AST.
        """
        source = self.desc.get_attr_by_id(align.source)
        target = self.desc.get_attr_by_id(align.target)

        to_aligned_dim = {
            step.target_idx: step.source_idx for step in align.aligned_steps
        }
        return PathAccessor(self.program).iterate_elements(
            ast,
            target,
            aligned_attr=source,
            to_aligned_dim=to_aligned_dim,
            validate_path=validate_path,
            on_missing_key=on_missing_key,
            iter_final_list=iter_final_list,
        )


class PathAccessor:
    """Generate code to access elements (indices & values) of an attribute"""

    def __init__(self, program: Program):
        self.program = program

    def iterate_elements(
        self,
        ast: AST,
        attr: Attr,
        aligned_attr: Optional[Attr] = None,
        to_aligned_dim: Optional[dict[int, int]] = None,
        validate_path: bool = False,
        on_missing_key: Optional[Callable[[AST], Any]] = None,
        iter_final_list: bool = False,
        on_step_callback: Optional[OnStepCallback] = None,
    ):
        ast = ast.update_recursively(
            fn=lambda ast, dim: self.next_dimensions(
                ast,
                attr,
                dim,
                aligned_attr,
                to_aligned_dim,
                validate_path,
                on_missing_key,
                iter_final_list,
                on_step_callback,
            ),
            context=0,
        )
        return ast

    def next_dimensions(
        self,
        ast: AST,
        attr: Attr,
        dim: int,
        aligned_attr: Optional[Attr],
        to_aligned_dim: Optional[dict[int, int]],
        validate_path: bool,
        on_missing_key: Optional[Callable[[AST], Any]] = None,
        iter_final_list: bool = False,
        on_step_callback: Optional[OnStepCallback] = None,
    ):
        """Generate code to access elements of dimensions of attr started at dim.
        Return the next ast, remaining dimension index, and whether it has stopped.

        Args:
            mem: memory to store variables
            ast: current ast
            attr: attribute
            dim: starting dimension
            aligned_attr:
            to_aligned_dim:
            validate_path: whether to generate code to check if values of each dimension is correct. If the attribute
                is annotated with `missing_path = True`, then setting this does not have any effect.
            on_missing_key: a function that will be called when the key does not exist. If it is None, then we will raise an exception.
            iter_final_list: if value type of attribute is a list, and this flag is true, we will iterate over the list to yield each item
            on_step_callback: a function that can be called to generate code at each time we step into a new dimension (except dimension 0, which is the resource data)
        """
        n_dim = len(attr.path.steps)
        if attr.value_type.is_list() and iter_final_list:
            n_dim += 1

        if dim >= n_dim:
            return ast, dim, True

        if dim == 0:
            collection = self.program.get_var(
                key=VarSpace.resource_data(attr.resource_id), at=ast.next_child_id()
            )
        else:
            collection = self.program.get_var(
                key=VarSpace.attr_value_dim(attr.resource_id, attr.id, dim - 1),
                at=ast.next_child_id(),
            )

        if dim == len(attr.path.steps):
            # we are iterating over the value list
            step = path.RangeExpr(0, None, 1)
        else:
            step = attr.path.steps[dim]

        # index expr does not need nested ast.
        while isinstance(step, path.IndexExpr) and dim < len(attr.path.steps):
            if isinstance(step.val, path.Expr):
                # I don't know about recursive path expression yet -- what usecase and how to use them -- so I can't implement.
                raise Exception(
                    f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.val}"
                )

            # we do not need nested loop for index expression as we can just directly access the value
            c1 = DeferredVar(
                name=f"{attr.id}_value_{dim}",
                key=VarSpace.attr_value_dim(attr.resource_id, attr.id, dim),
            )

            if validate_path and not attr.path.is_step_optional(dim):
                handle_missing_key = "safe"
            elif attr.path.is_step_optional(dim):
                assert on_missing_key is not None
                handle_missing_key = on_missing_key
            else:
                handle_missing_key = "no_missing_key"

            ast = self.access_key(
                ast,
                attr,
                expr.ExprVar(collection),
                expr.ExprConstant(step.val),
                c1,
                dim,
                handle_missing_key,
                on_step_callback,
            )

            collection = c1.get_var()
            dim += 1
            if dim == len(attr.path.steps):
                # we have reached the end of the path
                # however, if the value type is a list, and we enable iter_final_list,
                # we need the final iteration
                return ast, dim, not (attr.value_type.is_list() and iter_final_list)
            step = attr.path.steps[dim]

        assert not isinstance(step, path.IndexExpr), (attr, step, dim)

        # other exprs require nested statement (for loop)

        if isinstance(step, path.RangeExpr):
            itemindex = DeferredVar(
                name=f"{attr.id}_index_{dim}",
                key=VarSpace.attr_index_dim(attr.resource_id, attr.id, dim),
            )
            itemvalue = DeferredVar(
                name=f"{attr.id}_value_{dim}",
                key=VarSpace.attr_value_dim(attr.resource_id, attr.id, dim),
            )

            if to_aligned_dim is not None and dim in to_aligned_dim:
                # this attribute has been aligned with other attribute and
                # the dimension is bound to the previously set dimension (of a subject)
                # so we need to copy the value
                assert aligned_attr is not None
                aligned_dim = to_aligned_dim[dim]
                aligned_dim_index = self.program.get_var(
                    key=VarSpace.attr_index_dim(
                        aligned_attr.resource_id, aligned_attr.id, aligned_dim
                    ),
                    at=ast.next_child_id(),
                )

                if step == aligned_attr.path.steps[aligned_dim]:
                    # now if the start, end, and step between the two attrs are the same, we just copy the value
                    # otherwise, we need to readjust the index
                    ast.assign(itemindex, expr.ExprVar(aligned_dim_index))
                else:
                    # recalculate the index
                    raise NotImplementedError()

                itemindex = itemindex.get_var()
                if validate_path:
                    invok_item_getter = DReprPredefinedFn.safe_item_getter(
                        self.program,
                        expr.ExprVar(collection),
                        expr.ExprVar(itemindex),
                        expr.ExprConstant(
                            f"Encounter error while accessing element of attribute {attr.id} via alignment at "
                            + f"dimension = {dim} - full path = {attr.path.to_lang_format()}"
                        ),
                    )
                else:
                    invok_item_getter = PredefinedFn.item_getter(
                        expr.ExprVar(collection), expr.ExprVar(itemindex)
                    )

                ast.assign(itemvalue, invok_item_getter)
                itemvalue = itemvalue.get_var()
                if on_step_callback is not None:
                    on_step_callback(
                        ast,
                        dim,
                        expr.ExprVar(collection),
                        expr.ExprVar(itemindex),
                        itemvalue,
                    )
            else:
                # the dimension is not bound, we are going to generate multiple values
                # using a for loop
                start_var = DeferredVar(
                    name=f"start",
                    key=("local-var", "start", f"attr={attr.id}", f"ast={ast.id}"),
                )
                if isinstance(step.start, path.Expr):
                    # I don't know about recursive path expression yet.
                    raise Exception(
                        f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.start}"
                    )
                ast.assign(start_var, expr.ExprConstant(step.start))
                start_var = start_var.get_var()

                end_var = DeferredVar(
                    name=f"end",
                    key=("local-var", "end", f"attr={attr.id}", f"ast={ast.id}"),
                )
                if step.end is None:
                    if validate_path:
                        invok_len = DReprPredefinedFn.safe_len(
                            self.program,
                            expr.ExprVar(collection),
                            expr.ExprConstant(
                                f"Encounter error while computing number of elements of attribute {attr.id} at "
                                + f"dimension = {dim} - full path = {attr.path.to_lang_format()}"
                            ),
                        )
                    else:
                        invok_len = PredefinedFn.len(expr.ExprVar(collection))
                    ast.assign(end_var, invok_len)
                else:
                    if isinstance(step.end, path.Expr):
                        # I don't know about recursive path expression yet.
                        raise Exception(
                            f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.end}"
                        )

                    if validate_path:
                        invok_len = DReprPredefinedFn.safe_len(
                            self.program,
                            expr.ExprVar(collection),
                            expr.ExprConstant(
                                f"Encounter error while computing number of elements of attribute {attr.id} at "
                                + f"dimension = {dim} - full path = {attr.path.to_lang_format()}"
                            ),
                        )

                        ast.if_(
                            expr.ExprLessThanOrEqual(
                                invok_len, expr.ExprConstant(step.end)
                            )
                        )(
                            stmt.ExceptionStatement(
                                PredefinedFn.base_error(
                                    f"Trying to select out-of-bound elements of attribute {attr.id} at dimension = {dim} - full path = {attr.path.to_lang_format()}"
                                )
                            )
                        )

                    ast.assign(end_var, expr.ExprConstant(step.end))
                end_var = end_var.get_var()

                if isinstance(step.step, path.Expr):
                    # I don't know about recursive path expression yet.
                    raise Exception(
                        f"Recursive path expression is not supported yet. Please raise a ticket to notify us for future support! Found: {step.step}"
                    )
                elif step.step != 1:
                    step_var = DeferredVar(f"step__local_ast_{ast.id}")
                    ast.assign(step_var, expr.ExprConstant(step.step))
                    step_var = step_var.get_var()
                    expr_step_var = expr.ExprVar(step_var)
                else:
                    expr_step_var = None

                ast = ast.for_loop(
                    item=itemindex,
                    iter=PredefinedFn.range(
                        expr.ExprVar(start_var), expr.ExprVar(end_var), expr_step_var
                    ),
                )
                itemindex = itemindex.get_var()

                ast.assign(
                    itemvalue,
                    PredefinedFn.item_getter(
                        expr.ExprVar(collection), expr.ExprVar(itemindex)
                    ),
                )
                itemvalue = itemvalue.get_var()
                if on_step_callback is not None:
                    on_step_callback(
                        ast,
                        dim,
                        expr.ExprVar(collection),
                        expr.ExprVar(itemindex),
                        itemvalue,
                    )
            return (
                ast,
                dim + 1,
                False,
            )

        raise NotImplementedError(step)

    def access_key(
        self,
        ast: AST,
        attr: Attr,
        collection: expr.Expr,
        key: expr.ExprConstant,
        result: DeferredVar | Var,
        dim: int,
        handle_missing_key: Literal["safe", "no_missing_key"] | Callable[[AST], None],
        on_step_callback: Optional[OnStepCallback] = None,
    ):
        if handle_missing_key == "no_missing_key":
            ast.assign(result, PredefinedFn.item_getter(collection, key))
            if on_step_callback is not None:
                on_step_callback(ast, dim, collection, key, result)
            return ast

        if handle_missing_key == "safe":
            ast.assign(
                result,
                DReprPredefinedFn.safe_item_getter(
                    self.program,
                    collection,
                    key,
                    expr.ExprConstant(
                        f"While traveling elements of attribute {attr.id}, encounter key error: "
                        + f"key {key.constant} does not exist (key position = {dim} - full path = {attr.path.to_lang_format()})"
                    ),
                ),
            )
            if on_step_callback is not None:
                on_step_callback(ast, dim, collection, key, result)
            return ast

        ast.if_(expr.ExprNegation(PredefinedFn.has_item(collection, key)))(
            # if the key does not exist, we call the function to handle it
            # because we do not step in -- we don't invoke the callback here
            handle_missing_key
        )
        inner_ast = ast.else_()
        inner_ast.assign(result, PredefinedFn.item_getter(collection, key))
        if on_step_callback is not None:
            on_step_callback(inner_ast, dim, collection, key, result)
        return inner_ast


class OnStepCallback(Protocol):
    def __call__(
        self,
        ast: AST,
        dim: int,
        collection: expr.Expr,
        key: expr.Expr,
        result: DeferredVar | Var,
    ): ...
